'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import typing
from collections import namedtuple
from datetime import datetime, date
from itertools import filterfalse
from addict import Dict


from cybernetickit.util import is_attr_required

from elasticsearch_dsl import Document, Date, Text, Keyword, Integer


def simple_mapping(anno):
    if anno == str:
        return Keyword()
    elif anno == int:
        return Integer()
    else:
        return Text()


class ___AliasDocType(Document):

    pk: str = Text(fields={'raw': Keyword()})
    inserted_at: date = Date()
    model = Dict({'_meta': Dict({'object_name': ""})})
    _meta = Dict({'_meta': Dict({'object_name': ""})})


    @classmethod
    def alias(cls):
        return cls.Index.name

    @classmethod
    def setup(cls, alias, pattern):
        index_template = cls._index.as_template(alias, pattern)
        index_template.save()
        # if not cls._index.exists():
        #    cls.migrate(alias, pattern, move_data=False)

    @classmethod
    def migrate(cls, alias, pattern, move_data=True, update_alias=True):
        next_index = pattern.replace('*', datetime.now().strftime('%Y.%m.%d'))
        es = connections.get_connection()
        es.indices.create(index=next_index)
        if move_data:
            # move data from current alias to the new index
            es.reindex(
                body={"source": {"index": alias}, "dest": {"index": next_index}},
                request_timeout=3600
            )
            # refresh the index to make the changes visible
            es.indices.refresh(index=next_index)
        if update_alias:
            # repoint the alias to point to the newly created index
            es.indices.update_aliases(body={
                'actions': [
                    {"remove": {"alias": alias, "index": pattern}},
                    {"add": {"alias": alias, "index": next_index}},
                ]
            })


ElasticModel = ___AliasDocType

class CoreModel(___AliasDocType):
    def __init__(self, *args, **kwargs):
        hints = typing.get_type_hints(type(self))
        required = filterfalse(lambda key: is_attr_required(hints[key]), hints.keys())
        namedtuple("klass", required)(**kwargs)
        return super(CoreModel, self).__init__(*args, **kwargs)



from elasticsearch_dsl import Keyword, Date, Boolean, Text
from cybernetickit import CoreModel


class OSFWork(CoreModel):
    title = Text()
    name = Keyword()
    description = Text()
    affiliations = Text(multi=True)
    contributors = Text(multi=True)
    date = Date()
    date_created = Date()
    date_modified = Date()
    date_published = Date()
    date_updated = Date()
    funders = Text(multi=True)
    hosts = Keyword(multi=True)
    identifiers = Keyword(multi=True)
    justification = Text()
    language = Keyword()
    publishers = Keyword(multi=True)
    retracted = Boolean()
    sources = Keyword(multi=True)
    subjects = Keyword(multi=True)
    tags = Keyword(multi=True)
    types = Keyword(multi=True)
    registration_type = Keyword()
    withdrawn = Boolean()
    catalog = Keyword()

    class Meta:
        index = 'osfwork'
        doc_type = 'osfwork'

    class Index:
        name = 'osfwork'



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from collections import namedtuple
from importlib import import_module

current_module = import_module('.constants', package='cybernetickit')


POLLUTANTS = [
    {'name': 'nitric oxide', 'parameter': 'NO', 'units': 'PPB'},
    {'name': 'nitrogen oxides', 'parameter': 'NOX', 'units': 'PPB'},
    {'name': 'total reactive nitrogen', 'parameter': 'NOY', 'units': 'PPB'},
    {'name': 'nitrogen dioxide', 'parameter': 'NO2T', 'units': 'PPB'},
    {'name': 'nitrogen dioxide, computed', 'parameter': 'NO2', 'units': 'PPB'},
    {'name': 'nitrogen dioxide, computed', 'parameter': 'NO2Y', 'units': 'PPB'},
    {'name': 'nitrate, not adjusted for ammonium ion', 'parameter': 'NO3', 'units': 'UG/M3'},
    {'name': 'sulfate, not adjusted for ammonium ion', 'parameter': 'SO4', 'units': 'UG/M3'},
    {'name': 'sulfur dioxide, conventional', 'parameter': 'SO2', 'units': 'PPB'},
    {'name': 'sulfur dioxide, 24-hr average', 'parameter': 'SO2-24HR', 'units': 'PPB'},
    {'name': 'sulfur dioxide, trace levels', 'parameter': 'SO2T', 'units': 'PPB'},
    {'name': 'carbon monoxide, conventional', 'parameter': 'CO', 'units': 'PPM'},
    {'name': 'carbon monoxide, 8-hr average', 'parameter': 'CO-8HR', 'units': 'PPM'},
    {'name': 'carbon monoxide, trace levels', 'parameter': 'COT', 'units': 'PPM'},
    {'name': 'elemental carbon, PM2.5', 'parameter': 'EC', 'units': 'UG/M3'},
    {'name': 'organic carbon, not adjusted for oxygen and hydrogen, PM2.5', 'parameter': 'OC', 'units': 'UG/M3'},
    {'name': 'black carbon at 880 nm', 'parameter': 'BC', 'units': 'UG/M3'},
    {'name': 'second channel of Aethalometer at 370 nm', 'parameter': 'UV-AETH', 'units': 'UG/M3'},
    {'name': 'PM2.5 mass', 'parameter': 'PM2.5', 'units': 'UG/M3'},
    {'name': 'PM10 mass', 'parameter': 'PM10', 'units': 'UG/M3'},
    {'name': 'PM2.5 mass 24-hr average', 'parameter': 'PM2.5-24HR', 'units': 'UG/M3'},
    {'name': 'PM10 mass 24-hr average', 'parameter': 'PM10-24HR', 'units': 'UG/M3'},
    {'name': 'ozone', 'parameter': 'OZONE', 'units': 'PPB'},
    {'name': 'peak ozone 8-hr average', 'parameter': 'OZONE-8HR', 'units': 'PPB'},
    {'name': 'peak ozone 1-hr maxmium', 'parameter': 'OZONE-1HR', 'units': 'PPB'},
    {'name': 'ambient temperature', 'parameter': 'TEMP', 'units': 'C'},
    {'name': 'wind speed', 'parameter': 'WS', 'units': 'M/S'},
    {'name': 'wind direction', 'parameter': 'WD', 'units': 'DEGREES'},
    {'name': 'relative humidity', 'parameter': 'RHUM', 'units': 'PERCENT'},
    {'name': 'barometric pressure', 'parameter': 'BARPR', 'units': 'MILLIBAR'},
    {'name': 'solar radiation', 'parameter': 'SRAD', 'units': 'WATTS/M2'},
    {'name': 'precipitation', 'parameter': 'PRECIP', 'units': 'MM'},
]


for p in POLLUTANTS:
    name = p.get('parameter').replace('-', '_').replace('.', '')
    fields = ['name', 'parameter', 'units']
    setup = namedtuple(name, fields)

    setattr(current_module, name, setup(parameter=p.get('parameter'),
                                        name=p.get('name'),
                                        units=p.get('units')))



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from abc import ABCMeta, abstractmethod
from typing import Any

from cybernetickit import CoreModel


class Repository(metaclass=ABCMeta):

    def __init__(self, model: CoreModel):
        pass

    @abstractmethod
    def create(self, model: CoreModel) -> CoreModel:
        model.save()

    @abstractmethod
    def get(self, id: Any) -> CoreModel:
        return model.get(id)

    @abstractmethod
    def delete(self, id: Any) -> CoreModel:
        pass

    @abstractmethod
    def update(self, model: CoreModel) -> CoreModel:
        pass

    @abstractmethod
    def update(self, id: Any, fields: dict) -> CoreModel:
        pass



from elasticsearch_dsl import Keyword, GeoShape, GeoPoint, Object, Text

from cybernetickit import CoreModel, geo


class County(CoreModel):
    name = Text()
    name_full = Text()
    fips = Keyword()
    state = Keyword()
    geometry = GeoShape()
    centroid = GeoPoint()
    area = Object(geo.Area)

    class Index:
        name = 'county'
        doc_type = 'county'

    class Meta:
        doc_type = 'county'



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from elasticsearch_dsl import Keyword, Nested, Date, Object, Text
from cybernetickit import CoreModel
from typing import List, Optional


class Preprint(CoreModel):
    title: str = Text()
    description: str = Text(multi=True)
    identifier: str = Keyword(multi=True)
    oai_id: str = Keyword()
    publisher: str = Keyword()
    set: List[dict] = Object(properties={
                 'code': Keyword(),
                 'name': Keyword()
                 }, multi=True)
    creator: List[str] = Text(multi=True)
    subject: List[str] = Keyword(multi=True)
    type: str = Keyword()
    date: List[str] = Date(multi=True)
    uploaded_at = Date()
    format: str = Keyword()
    source: str = Keyword()
    language: str = Keyword()
    rights: Optional[List[str]] = Keyword(multi=True)
    repository: dict = Object(properties={
                        'id': Keyword(),
                        'name': Keyword()})

    class Meta:
        index = 'preprint'
        doc_type = 'preprint'

    class Index:
        name = 'preprint'


class Repository(CoreModel):
    name: str = Text()
    description: str = Text()
    topic: str = Keyword()
    type: str = Keyword()
    tags: List[str] = Keyword(multi=True)
    city: Optional[str] = Keyword()  # blank for state and federal datasets
    state: Optional[str] = Keyword()  # blank for federal datasets
    administrative_level: str = Keyword()  # city, state, federal
    institution: str = Text()
    agency: str = Text()
    base_url: str = Keyword()
    formats = Nested(properties={
                        'name': Keyword(),
                        'schema': Keyword(),
                        'url': Keyword()
                        })
    version: str = Keyword()
    emails: List[str] = Keyword(multi=True)

    sets = Nested(properties={
                        'namespace': Keyword(),
                        'name': Keyword()
                    })

    class Meta:
        index = 'repository'
        doc_type = 'repository'

    class Index:
        name = 'repository'



'''
Copyright (C) 2021  Civic Hacker, LLC <opensource@civichacker.com>

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
'''

import cmath


def latlon_to_complex(lat=0.0, lon=0.0):
    return complex((cmath.tan(lat - cmath.tau/4) + cmath.sin(lon)), cmath.cos(lon))



from elasticsearch_dsl import Keyword, GeoShape, GeoPoint, Object, Boolean

from cybernetickit import CoreModel, geo


class Tract(CoreModel):
    name = Keyword()
    name_full = Keyword()
    fips = Keyword()
    ce = Keyword()
    state = Keyword()
    county = Keyword()
    geometry = GeoShape()
    centroid = GeoPoint()
    area = Object(geo.Area)
    opportunityzone = Boolean()

    class Index:
        name = 'tract'
        doc_type = 'tract'

    class Meta:
        doc_type = 'tract'



from elasticsearch_dsl import Keyword, Integer, Object

from cybernetickit import CoreModel as AliasDocType
from cybernetickit.county import County
from cybernetickit.state import State


class CensusBlock(AliasDocType):
    fips = Keyword()
    state = Object(State)
    county = Object(County)
    tract = Keyword()

    # Totals
    total = Integer()  # P0120001 also P0040001
    male = Integer()  # P0120002
    female = Integer()  # P0120026

    non_hispanic = Integer()  # P0040002
    hispanic = Integer()  # P0040003

    # Non-Hispanic Race Totals
    white = Integer()  # P0070003
    black = Integer()  # P0070004
    native = Integer()  # P0070005
    asian = Integer()  # P0070006
    hawaiian = Integer()  # P0070007
    other = Integer()  # P0070008

    # Hispanic Race Totals
    hispanic_white = Integer()  # P0070010
    hispanic_black = Integer()  # P0070011
    hispanic_native = Integer()  # P0070012
    hispanic_asian = Integer()  # P0070013
    hispanic_hawaiian = Integer()  # P0070014
    hispanic_other = Integer()  # P0070015

    # Race-Gender Breakdowns
    white_male = Integer()  # P012A002 or P012A025
    white_female = Integer()  # P012A026 or P012A049
    black_male = Integer()  # P012B002
    black_female = Integer()  # P012B026
    native_male = Integer()  # P012C002
    native_female = Integer()  # P012C026
    asian_male = Integer()  # P012D002
    asian_female = Integer()  # P012D026
    hawaiian_male = Integer()  # P012E002
    hawaiian_female = Integer()  # P012E026
    other_male = Integer()  # P012F002
    other_female = Integer()  # P012F026

    class Index:
        name = 'block'
        doc_type = 'block'

    class Meta:
        doc_type = 'block'



from typing import List, Any
from elasticsearch_dsl import Text, Keyword, Object, InnerDoc, Date

from cybernetickit import CoreModel


class RequiredCommonProperties(InnerDoc):
    type: str = Keyword() # From STIX 2.1: type=='report'
    spec_version: str = Keyword()
    id: str = Keyword()
    created: Any = Keyword()
    modified: Any = Keyword()



class Source(InnerDoc):
    name: str = Text()
    url: str = Keyword()


class Report(CoreModel):
    stix_rcp = Object(RequiredCommonProperties)
    name: str = Text()
    description: str = Text()
    report_types: List[str] = Keyword(multi=True) # From STIX 2.1 report-type-ov + ours
    published = Date()
    object_refs = Keyword(multi=True)
    uid: str = Keyword()
    url: str = Keyword()
    source: Any = Object(Source)
    locality: List[str] = Keyword(multi=True) # Must match datasource's locality vocab
    last_revision = Date()

    class Index:
        name = 'stix'



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC
This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.
You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from typing import List
from datetime import datetime, date
from elasticsearch_dsl import Keyword, Nested, Date, Object, Text, InnerDoc, GeoPoint, Integer
from cybernetickit import CoreModel


REPO_TYPES = [
    {'id': 'socrata', 'name': 'Socrata'},
    {'id': 'ckan', 'name': 'CKAN'},
    {'id': 'dkan', 'name': 'DKAN'},
    {'id': 'opendatasoft', 'name': 'OpenDataSoft'},
    {'id': 'oai', 'name': 'OpenArchives Initiative'},
    {'id': 'arcgis', 'name': 'ArcGIS'},
    {'id': 'web', 'name': 'Generic Web Datasource'},
    {'id': 'frictionless', 'name': 'Frictionless Data'},
    {'id': 'junar', 'name': 'Junar'},
    {'id': 'google', 'name': 'Google Datasource'}
]

class DataSourceType(InnerDoc):
    name: str = Keyword()
    identifier: str = Keyword()


class Repository(InnerDoc):
    name: str = Keyword()
    id: str = Keyword()
    version: str = Keyword()
    url: str = Keyword()


class DataHighlight(InnerDoc):
    name: str = Keyword()
    id: str = Keyword()
    version: str = Keyword()
    url: str = Keyword()
    modified: date = Date()
    created: date = Date()


class DataSource(CoreModel):
    id: str = Keyword()
    url = Keyword()
    name: str = Text()
    level: str = Keyword()  # city, state, federal
    locality: List[str] = Keyword(multi=True)
    total: int = Integer()
    new = Object(DataHighlight)
    updated = Object(DataHighlight)
    location = GeoPoint()
    version: str = Keyword()
    type = Object(DataSourceType)

    class Index:
        name = 'datasource'
        doc_type = 'datasource'

    class Meta:
        doc_type = 'datasource'

    def add_updated(self, name, url, identifier, created, updated):
        self.updated.append(
            DataHighlight(name=name, url=url, id=identifier, created=created, updated=updated)
        )

    def add_new(self, name, url, identifier, created, updated=None):
        self.new.append(
            DataHighlight(name=name, url=url, id=identifier, created=created, updated=updated)
        )


class DataSourceDoc(InnerDoc):
    id: str = Keyword()
    url: str = Keyword()
    name: str = Text()
    level: str = Keyword()  # city, state, federal
    locality: List[str] = Keyword(multi=True)
    location = GeoPoint()
    version: str = Keyword()
    type = Object(DataSourceType)

    class Index:
        name = 'datasource'
        doc_type = 'datasource'

    class Meta:
        doc_type = 'datasource'


class PointOfContact(InnerDoc):
    fn: str = Keyword()
    ln: str = Keyword()
    email: str = Text()


class DataSet(CoreModel):

    class Index:
        name = 'dataset'
        doc_type = 'dataset'

    class Meta:
        doc_type = 'dataset'

    name: str = Text()
    format: str = Keyword()
    description: str = Text()
    category: str = Keyword()
    type: str = Keyword()
    repository = Object(Repository)
    tags: List[str] = Keyword(multi=True)
    raw = Object()
    source = Object(DataSourceDoc)
    url: str = Keyword()
    odata: str = Keyword()
    identifier: str = Keyword()
    landing: str = Keyword()
    distribution = Object(properties={
        '@type': Keyword(),
        'downloadURL': Keyword(),
        'mediaType': Keyword(),
    })
    modified: date  = Date()
    created: date = Date()
    locality: List[str] = Keyword(multi=True)
    level: str = Keyword()
    contact = Object(PointOfContact)

    def add_distribution(self, download):
        self.distribution.append({
            '@type': download['@type'],
            'downloadURL': download['downloadURL'],
            'mediaType': download['mediaType']
        })


class Data(CoreModel):
    locality: str = Keyword(multi=True)
    level: str = Keyword()
    identifier: str = Keyword()
    field = Object(properties={
        'name': Keyword(),
        'type': Keyword()})

    class Index:
        name = 'data'
        doc_type = 'data'



from elasticsearch_dsl import GeoShape, Keyword, GeoPoint, Object

from cybernetickit import CoreModel, geo


class State(CoreModel):
    name = Keyword()
    abbr = Keyword()
    fips = Keyword()
    geometry = GeoShape()
    area = Object(geo.Area)
    centroid = GeoPoint()

    class Index:
        name = 'state'
        doc_type = 'state'

    class Meta:
        doc_type = 'state'



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import sys
import typing

if sys.version_info.major == 3 and sys.version_info.minor == 7:
    import typing_inspect
    def is_attr_required(atype) -> bool:
        return typing_inspect.get_origin(atype) is typing.Union and type(None) in typing_inspect.get_args(atype)
else:

    def is_attr_required(atype) -> bool:
        return typing.get_origin(atype) is typing.Union and type(None) in typing.get_args(atype)



'''
Copyright (C) 2021  Civic Hacker, LLC <opensource@civichacker.com>

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
'''

from datetime import datetime
from enum import Enum
from typing import Optional

from elasticsearch_dsl import Date, Text, Keyword, Integer

from cybernetickit import CoreModel

class DispositionEnum(str, Enum):
    enforced = 'enforce'
    report = 'report'


class CSPViolationReport(CoreModel):
    created_at: datetime = Date()
    document_uri: str = Text()
    referrer: str = Text()
    disposition: DispositionEnum = Keyword()
    effective_directive: str = Text()
    original_policy: str = Text()
    script_sample: str = Text()
    status_code: int = Integer()
    violated_directive: str = Text()
    line_number: int = Integer()
    client_ip: str = Keyword()

    class Index:
        name = "csp-*"

    async def save(self, **kwargs):
        today = datetime.utcnow().strftime('%Y.%m.%d')
        self.meta.index = f'csp-{today}'
        return await super().save(**kwargs)



from elasticsearch_dsl import Keyword, Object, InnerDoc
from cybernetickit import CoreModel


REPO_TYPES = [
    {'id': 'socrata', 'name': 'Socrata'},
    {'id': 'ckan', 'name': 'CKAN'},
    {'id': 'dkan', 'name': 'DKAN'},
    {'id': 'opendatasoft', 'name': 'OpenDataSoft'},
    {'id': 'oai', 'name': 'OpenArchives Initiative'},
    {'id': 'arcgis', 'name': 'ArcGIS'},
    {'id': 'web', 'name': 'Generic Web Datasource'},
    {'id': 'frictionless', 'name': 'Frictionless Data'},
    {'id': 'google', 'name': 'Google Datasource'}
]


class VegaSpec(InnerDoc):
    encodings = Object(multi=True)
    mark = Keyword()
    autosize = Object()


class Vega(CoreModel):

    class Index:
        name = 'vega'
        doc_type = 'vega'

    class Meta:
        doc_type = 'vega'

    spec = Object(VegaSpec)
    data = Keyword()
    id = Keyword()
    url = Keyword()
    chooseBy = Keyword()



from elasticsearch_dsl import Keyword, GeoShape, GeoPoint, Nested, Object, Double, Integer, Text, InnerDoc
from cybernetickit import CoreModel, geo


class SubjectScore(InnerDoc):
    metric = Keyword()
    value = Double()
    type = Keyword()


class District(InnerDoc):
    code = Keyword()
    name = Text()
    state = Keyword()


class SchoolDistrictPercolate(geo.GeoPercolate):
    state = Keyword()

    class Index:
        name = 'perc'
        doc_type = 'perc'

    class Meta:
        doc_type = 'perc'


class SchoolDistrict(CoreModel):
    name = Keyword()
    state = Keyword()
    year = Keyword()
    geoid = Keyword()
    lowest_grade = Keyword()
    highest_grade = Keyword()
    geometry = GeoShape()
    centroid = GeoPoint()
    kind = Keyword()
    area = Object(geo.Area)

    class Index:
        name = 'school.district'
        doc_type = 'school.district'

    class Meta:
        doc_type = 'school.district'


class AssessmentScore(CoreModel):
    year = Keyword()
    state = Keyword()
    subgroup = Keyword()
    participation_rate = Double()
    total_tests = Integer()
    science = Object(SubjectScore)
    english = Object(SubjectScore)
    math = Object(SubjectScore)
    reading = Object(SubjectScore)
    composite = Object(SubjectScore)
    type = Keyword()
    district = Object(District)
    fact = Nested(properties={
        'name': Text(),
        'value': Double(),
        'type': Keyword()
    })

    class Index:
        name = 'assessment'
        doc_type = 'assessment'

    class Meta:
        doc_type = 'assessment'

    def add_fact(self, fact_title, fact_value, fact_type='numeric'):
        self.fact.append({
            'title': fact_title,
            'type': fact_type,
            'value': fact_value
        })



from elasticsearch_dsl import Keyword, Date, Text, Boolean
from cybernetickit import CoreModel


class Book(CoreModel):
    id = Keyword()
    isbn = Text()
    title = Text()
    website = Text()
    path = Text()
    date_added = Date()
    format = Keyword()
    deleted = Boolean(default=False)
    filename = Text()
    category = Text()
    tags = Keyword(multi=True)
    level = Keyword()

    class Index:
        name = 'book'

    class Meta:
        doc_type = 'book'



'''
Copyright (C) 2021  Civic Hacker, LLC <opensource@civichacker.com>

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<http://www.gnu.org/licenses/>.
'''


from elasticsearch_dsl import Keyword, GeoShape, GeoPoint, Text, Double
from elasticsearch_dsl import Object, InnerDoc, Percolator
from cybernetickit import CoreModel
from cybernetickit import CoreModel as AliasDocType


class Area(InnerDoc):
    land: float = Double()
    water: float = Double()


class State(CoreModel):
    name: str = Keyword()
    abbr: str = Keyword()
    fips: str = Keyword()
    geometry = GeoShape()
    area = Object(Area)
    centroid = GeoPoint()

    class Index:
        name = 'state'
        doc_type = 'state'

    class Meta:
        doc_type = 'state'


class County(CoreModel):
    name: str = Text()
    name_full: str = Text()
    fips: str = Keyword()
    state: str = Keyword()
    geometry = GeoShape()
    centroid = GeoPoint()
    area = Object(Area)

    class Index:
        name = 'county'
        doc_type = 'county'

    class Meta:
        doc_type = 'county'


class GeoPercolate(AliasDocType):
    geometry = GeoShape()
    query = Percolator()

    class Index:
        name = 'perc'
        doc_type = 'perc'

    class Meta:
        doc_type = 'perc'



'''
Cybernetic Kit
Copyright (C) 2021  Civic Hacker, LLC

This program is free software: you can redistribute it and/or modify
it under the terms of the Lesser GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Lesser GNU General Public License for more details.

You should have received a copy of the Lesser GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from datetime import date, datetime, time
from typing import List

from elasticsearch_dsl import Object, Document, InnerDoc, GeoShape, Percolator
from elasticsearch_dsl import Keyword, Float, Date, Text, Integer, GeoPoint

from cybernetickit import CoreModel, geo

POLLUTANTS = [
    {'name': 'nitric oxide', 'parameter': 'NO', 'units': 'PPB'},
    {'name': 'nitrogen oxides', 'parameter': 'NOX', 'units': 'PPB'},
    {'name': 'total reactive nitrogen', 'parameter': 'NOY', 'units': 'PPB'},
    {'name': 'nitrogen dioxide', 'parameter': 'NO2T', 'units': 'PPB'},
    {'name': 'nitrogen dioxide, computed', 'parameter': 'NO2', 'units': 'PPB'},
    {'name': 'nitrogen dioxide, computed', 'parameter': 'NO2Y', 'units': 'PPB'},
    {'name': 'nitrate, not adjusted for ammonium ion', 'parameter': 'NO3', 'units': 'UG/M3'},
    {'name': 'sulfate, not adjusted for ammonium ion', 'parameter': 'SO4', 'units': 'UG/M3'},
    {'name': 'sulfur dioxide, conventional', 'parameter': 'SO2', 'units': 'PPB'},
    {'name': 'sulfur dioxide, 24-hr average', 'parameter': 'SO2-24HR', 'units': 'PPB'},
    {'name': 'sulfur dioxide, trace levels', 'parameter': 'SO2T', 'units': 'PPB'},
    {'name': 'carbon monoxide, conventional', 'parameter': 'CO', 'units': 'PPM'},
    {'name': 'carbon monoxide, 8-hr average', 'parameter': 'CO-8HR', 'units': 'PPM'},
    {'name': 'carbon monoxide, trace levels', 'parameter': 'COT', 'units': 'PPM'},
    {'name': 'elemental carbon, PM2.5', 'parameter': 'EC', 'units': 'UG/M3'},
    {'name': 'organic carbon, not adjusted for oxygen and hydrogen, PM2.5', 'parameter': 'OC', 'units': 'UG/M3'},
    {'name': 'black carbon at 880 nm', 'parameter': 'BC', 'units': 'UG/M3'},
    {'name': 'second channel of Aethalometer at 370 nm', 'parameter': 'UV-AETH', 'units': 'UG/M3'},
    {'name': 'PM2.5 mass', 'parameter': 'PM2.5', 'units': 'UG/M3'},
    {'name': 'PM10 mass', 'parameter': 'PM10', 'units': 'UG/M3'},
    {'name': 'PM2.5 mass 24-hr average', 'parameter': 'PM2.5-24HR', 'units': 'UG/M3'},
    {'name': 'PM10 mass 24-hr average', 'parameter': 'PM10-24HR', 'units': 'UG/M3'},
    {'name': 'ozone', 'parameter': 'OZONE', 'units': 'PPB'},
    {'name': 'peak ozone 8-hr average', 'parameter': 'OZONE-8HR', 'units': 'PPB'},
    {'name': 'peak ozone 1-hr maxmium', 'parameter': 'OZONE-1HR', 'units': 'PPB'},
    {'name': 'ambient temperature', 'parameter': 'TEMP', 'units': 'C'},
    {'name': 'wind speed', 'parameter': 'WS', 'units': 'M/S'},
    {'name': 'wind direction', 'parameter': 'WD', 'units': 'DEGREES'},
    {'name': 'relative humidity', 'parameter': 'RHUM', 'units': 'PERCENT'},
    {'name': 'barometric pressure', 'parameter': 'BARPR', 'units': 'MILLIBAR'},
    {'name': 'solar radiation', 'parameter': 'SRAD', 'units': 'WATTS/M2'},
    {'name': 'precipitation', 'parameter': 'PRECIP', 'units': 'MM'},
]


class Agency(InnerDoc):
    name: str = Text()
    id: str = Keyword()


class Site(CoreModel):
    id: str = Keyword()
    code = Keyword()
    name = Text()
    website = Text()
    status = Keyword()
    type = Keyword()
    start_time = Date()
    end_time = Date()
    agency = Object(Agency)
    use = Keyword()
    outputs = Keyword(multi=True)
    inputs = Keyword(multi=True)
    region = Keyword()
    location = GeoPoint()
    elevation = Keyword()
    state = Object(geo.State)
    county = Object(geo.County)

    class Index:
        name = 'site'

    class Meta:
        doc_type = 'site'


class Pollutant(InnerDoc):
    parameter: str = Keyword()
    name: str = Text()
    units: str = Keyword()


class GeoPercolate(CoreModel):
    geometry = GeoShape()
    query = Percolator()

    class Index:
        name = 'perc'


class AirQualityPercolate(GeoPercolate):
    location = GeoPoint()
    sitename: str = Keyword()
    aqsid: str = Keyword()

    class Index:
        name = 'airquality-perc'
        doc_type = 'airquality-perc'

    class Meta:
        doc_type = 'airquality-perc'


class AirQuality(CoreModel):
    date: datetime = Date()
    time: str = Keyword()
    aqsid: str = Keyword()
    sitename: str = Text()
    site = Object(Site)
    offset: str = Keyword()
    pollutant = Object(Pollutant)
    units: str = Keyword()
    concentration: float = Float()
    datasource: str = Keyword()
    locality: List[str] = Keyword(multi=True)
    state: str = Keyword()
    county: str = Keyword()
    aqi: int = Integer()

    class Index:
        name = 'airquality'
        doc_type = 'airquality'

    class Meta:
        doc_type = 'airquality'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cybernetickit', 'cybernetickit.experimental']

package_data = \
{'': ['*']}

install_requires = \
['elasticsearch-dsl>=7.0.0,<8.0.0']

extras_require = \
{':python_version >= "3.7" and python_version < "3.8"': ['typing-inspect>=0.7.1,<0.8.0']}

setup_kwargs = {
    'name': 'cybernetickit',
    'version': '0.3.2b2',
    'description': 'The Cybernetic Kit is a set of conventions and techniques useful for making software-driven social impact.',
    'long_description': "# Cybernetic Kit\n\nThe Cybernetic Kit is a set of conventions and techniques useful for making software-driven social impact. This repository holds a collection of software components for multiple fields of study, such as:\n\n- Open Data & Civic Engagement\n- Environmental Science\n- Library Science\n- Software Development\n- Sociology\n- Neuroscience\n\n\n## Meet the Cast\n\n### Data Objects\n\nData Objects are Python classes designed to make it easier to interact with data. Using Data Objects to describe something of interest tells computers how to manipulate, archive, inspect, and exchange data.\n\n#### Built-In Data Objects\n\n- civic data\n- research preprints and archives\n- air quality sensor readings\n- geopolitical shapes (e.g., county, state)\n\n\n**How to use built-in Data Objects**\n\nHere's an example of how you'd go about storing air quality sensor readings in Elasticsearch.\n\n```python\nfrom datetime import datetime\n\nfrom elasticsearch_dsl import connections\nfrom cybernetickit.sensor import AirQuality\nfrom cybernetickit import constants\n\nconnections.create_connection(hosts=['localhost'], timeout=20)\n\nobservation = AirQuality(date=datetime.now(),\n                         state='ga',\n                         pollutant=constants.OZONE._asdict(),\n                         units=constants.OZONE.units,\n                         concentration=2.3)\n\nobservation.save()\n```\n\nYou can use the Elasticsearch DSL to search and aggregations against an existing Elasticsearch index:\n\n```python\nfrom elasticsearch_dsl import Q\ns = AirQuality.search(Q('match', state='ny') & Q('match', **{'pollutant.parameter': 'OZONE'}))\nresponse = s.execute()\n\nfor obs in response:\n    ...\n```\n\n#### Custom Data Objects\n\nSimply subclass the `CoreModel` and describe something of interest. Use annotations for runtime validation:\n\n```python\nfrom typing from List, Optional\nfrom cybernetickit from CoreModel, Keyword\n\n\nclass UseOfForce(CoreModel):\n    force_applied: str = Keyword()\n    deadly_force: Optional[str] = Keyword()\n    source_ref: str = Keyword()\n    target_ref: str = Keyword()\n\n```\n\n\n\n## Getting Started\n\n```sh\npip install cybernetickit\n```\n\n```python\nfrom cybernetickit import CoreModel\n```\n\n\n## Licensing\n\n    Cybernetic Kit\n    Copyright (C) 2021  Civic Hacker, LLC\n\n    This program is free software: you can redistribute it and/or modify\n    it under the terms of the Lesser GNU General Public License as published by\n    the Free Software Foundation, either version 3 of the License, or\n    (at your option) any later version.\n\n    This program is distributed in the hope that it will be useful,\n    but WITHOUT ANY WARRANTY; without even the implied warranty of\n    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n    Lesser GNU General Public License for more details.\n\n    You should have received a copy of the Lesser GNU General Public License\n    along with this program.  If not, see <https://www.gnu.org/licenses/>.\n",
    'author': 'Jurnell Cockhren',
    'author_email': 'jurnell@civichacker.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from elasticsearch_dsl import Object, InnerDoc
from elasticsearch_dsl import Keyword, Date, Integer, GeoPoint, Boolean
from cybernetickit import CoreModel


class Duration(InnerDoc):
    length = Integer()
    units = Keyword()  # hours, minutes, days


class Microaggression(CoreModel):
    type = Keyword()  # race, sexuality, gender, age, intersection
    racism = Boolean()
    sexism = Boolean()
    ageism = Boolean()
    colorism = Boolean()
    xenophobia = Boolean()
    antilgbtqi = Boolean()
    location = GeoPoint()
    at = Date(default_timezone='UTC')
    duration = Object(Duration)
    visibility = Keyword()  # private, public
    source = Keyword()  # online, person, organization
    reporter = Keyword()  # anonymous or user's email

    class Index:
        name = 'microaggression'
        doc_type = 'microaggression'

    class Meta:
        doc_type = 'microaggression'



# Copyright (C) 2021  Civic Hacker, LLC
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.


from typing import Tuple, List, NewType, TypeVar, Type
from typing import Union, Set
from dataclasses import dataclass, field
from cybernetickit import CoreModel
import math
import cmath
from decimal import Decimal
import operator


Location = TypeVar('Location', Tuple[int, int], complex)


class TopologicalError(Exception):
    def __init__(self, message):
        self.message = message


@dataclass(frozen=True)
class LineSegment:
    end: Location
    start: Location

    def __iter__(self):
        return iter([self.end, self.start])

    def __matmul__(self, other):
        return (self.start - other.start)*(self.end - other.end)

    def __abs__(self):
        return abs(self.end - self.start)


@dataclass(frozen=True)
class Vector(LineSegment):
    end: Location
    start: Location = complex(0, 0)

    def __matmul__(self, other):
        return (self.start - other.start)*(self.end - other.end)

    @property
    def real(self):
        return self.end.real

    @property
    def imag(self):
        return self.end.imag


class Bloom(type):
    remember: int # Band
    understand: int
    apply: int
    analyze: int
    evaluate: int
    create: int

    def __init_subclass__(cls, /, strategy, **kwargs):
        pass


@dataclass
class LearningExperience(metaclass=Bloom):
    ___points: Set[complex] = field(default_factory=set)
    software: Type[Vector] = Vector(end=complex(.5, 1))
    policy: Type[Vector] = Vector(end=complex(-.5, 1))
    data: Type[Vector] = Vector(end=complex(1, 0))

    @property
    def points(self) -> Set[complex]:
        return self.___points

    @points.setter
    def points(self, new_point: complex):
        if new_point in self:
            self.___points.add(new_point)
        else:
            raise TopologicalError("New point must be in interior")

    @property
    def area(self):
        p = self.perimeter/2
        return math.sqrt(p*(p-abs(self.software))*(p-abs(self.policy))*(p-abs(self.data)))

    @property
    def perimeter(self) -> Union[float, Decimal]:
        return abs(self.software) + abs(self.data) + abs(self.policy)

    def dot(self, vector1=None, vector2=None):
        return math.acos((vector1.real*vector2.real + vector1.imag*vector2.imag)/(abs(vector1)*abs(vector2)))

    def __len__(self):
        return len(self.points)

    def __contains__(self, other: complex):

        if any(map(lambda c: c == other,[complex(0,0), self.software.end, self.data.end])):
            # exclude vertices
            return False


        dx = other.real-self.software.end.real;
        dy = other.imag-self.software.end.imag;
        dx21 = self.software.end.real-self.data.end.real;
        dy12 = self.data.end.imag-self.software.end.imag;

        D = dy12*(self.data.start.real-self.software.real) + dx21*(self.data.start.imag-self.software.end.imag);
        s = dy12*dx + dx21*dy;
        t = (self.software.end.imag-self.data.start.imag)*dx + (self.data.start.real-self.software.end.real)*dy;
        if (D<0):
            return s<=0 and t<=0 and s+t>=D
        return s>=0 and t>=0 and s+t<=D



from elasticsearch_dsl import Text, Keyword, Date, Boolean
from elasticsearch_dsl import GeoShape, GeoPoint, Double
from cybernetickit import CoreModel as AliasDocType


class City(AliasDocType):
    name = Keyword()
    fips = Keyword()
    state = Keyword()
    url = Keyword()
    merged = Boolean()
    geometry = GeoShape()
    centroid = GeoPoint()
    area = Double()

    class Index:
        name = 'city'
        doc_type = 'city'

    class Meta:
        doc_type = 'city'


class Department(AliasDocType):
    name = Text()
    url = Keyword()
    locality = Keyword(multi=True)
    city = Keyword()
    state = Keyword()

    class Index:
        name = 'department'
        doc_type = 'department'

    class Meta:
        doc_type = 'department'


class Commission(AliasDocType):
    name = Keyword()
    url = Keyword()
    locality = Keyword(multi=True)
    city = Keyword()
    state = Keyword()

    class Index:
        name = 'commission'
        doc_type = 'commission'

    class Meta:
        doc_type = 'commission'


class Contract(AliasDocType):
    id = Keyword()
    contractor = Keyword()
    buyer = Keyword()
    department = Keyword()
    expiration = Date()
    city = Keyword()
    state = Keyword()

    class Index:
        name = 'contract'
        doc_type = 'contract'

    class Meta:
        doc_type = 'contract'
